/*
 * PortAudio analysis portion of the MIDI test module.
 *
 * See test.cpp for documentation.
 */

#include "pa_analysis.hpp"
#include <assert.h>
#include <iostream>
#undef NDEBUG // Don't let VC override the assertions in Release mode

using namespace std;

void initialize_analysis_data(AnalysisData& data)
{
   data.ref_channel = -1;
   int x;
   for (x = 0; x < PA_LATENCY_HISTOGRAM_BINS; ++x)
      data.latency[x] = 0;
   for (x = 0; x < PA_JITTER_HISTOGRAM_BINS; ++x)
      data.jitter[x] = 0;
   for (x = 0; x < PA_WIDTH_HISTOGRAM_BINS; ++x)
      data.ref_width[x] = 0;
   for (x = 0; x < PA_WIDTH_HISTOGRAM_BINS; ++x)
      data.test_width[x] = 0;

   data.out_of_range_latency = 0;
   data.out_of_range_jitter = 0;
   data.out_of_range_ref_width = 0;
   data.out_of_range_test_width = 0;

   data.max_latency = 0;
   data.max_jitter = 0;
   data.max_ref_width = 0;
   data.max_test_width = 0;

   data.sum = 0;
   data.sumSq = 0;
   data.sum_ref_width = 0;
   data.sum_test_width = 0;
   data.sumSq_ref_width = 0;
   data.sumSq_test_width = 0;

   data.num_left = 0;
   data.num_right = 0;

   data.frames = 0;

   data.last_burst_left = 0;
   data.last_burst_right = 0;

   data.in_burst_left = false;
   data.in_burst_right = false;

   data.cancel = false;

   for (x = 0; x < LATENCIES_TO_SAVE; ++x)
   {
      data.saved_latency_values[x] = 0;
      data.saved_latency_times[x] = 0;
   }
   data.saved_latencies_num = 0;
}

// We do this separately from initialize_analysis_data() because we need
// ..to know data.period before we can do it.
void initialize_analysis_data_saved_samples(AnalysisData& data)
{
   // to avoid round-off error, use formula:
   //  a * b / c = ((a/c) * b) + ((a%c) * b / c)
   data.period_in_frames = ((SAMPLE_RATE / 10000) * data.period)
      + ((SAMPLE_RATE % 10000) * data.period / 10000);
   data.saved_samples = new float[data.period_in_frames * PERIODS_TO_SAVE * 2];
   for (int x = 0; x < data.period_in_frames * PERIODS_TO_SAVE * 2; ++x)
      data.saved_samples[x] = 0;
   data.saved_samples_index = 0;
}

void writeout_saved_samples_data(AnalysisData& data, ostream& file) 
{
   int index = data.saved_samples_index;

   for (int i = 0; i < data.period_in_frames * PERIODS_TO_SAVE * 2; i += 2)
   {
      // wrapping circular buffer
      if (index >= data.period_in_frames * PERIODS_TO_SAVE * 2)
	 index = 0;
      float left = data.saved_samples[index++];
      float right = data.saved_samples[index++];
      file << left << " " << right << endl;
   }
}

void writeout_saved_latencies_data(AnalysisData& data, ostream& file)
{
   for (int i = 0; i < data.saved_latencies_num; ++i)
   {
      file << data.saved_latency_times[i] << " "
	   << data.saved_latency_values[i]
	   << endl;
   }
}

/* Analysis callback (registered with PortAudio) */
int analysis_callback(void *inputBuffer, void *outputBuffer,
		      unsigned long framesPerBuffer,
		      PaTimestamp outTime, void *userData)
{
   // avoid unused variable warnings
   (void) outputBuffer;
   (void) outTime;

   // (real callback starts here)

   assert(userData != NULL);
   AnalysisData* data = (AnalysisData*)userData;

   // if we're supposed to cancel, and just waiting for main driver to notice,
   // ..don't do any further analysis
   if (data->cancel)
      return 0;

   if (inputBuffer != NULL)
   {
      float* input = (float*)inputBuffer;

      for (unsigned int x = 0; x < framesPerBuffer; ++x)
      {
	 const unsigned int frame = data->frames + x;

	 // add current frame to circular buffer
	 if (data->saved_samples_index
	     >= data->period_in_frames * PERIODS_TO_SAVE * 2)
	    data->saved_samples_index = 0;
	 data->saved_samples[data->saved_samples_index++] = *input;
	 data->saved_samples[data->saved_samples_index++] = *(input + 1);

	 // single-msg processing code
	 if (data->msgs == 1)
	 {
	    // left
	    
	    // only process if enough frames since the last burst start,
	    // (and if this is actually a burst, of course)
	    if ((data->num_left == 0
		 || (frame - data->last_burst_left) > SKIP_MARGIN * data->msg_duration)
		&& (*input > THRESHOLD || *input < -THRESHOLD))
	    {
	       data->last_burst_left = frame;
	       // if this was the first burst, it's the reference channel
	       if (data->ref_channel == -1)
		  data->ref_channel = 0;
	       
	       if (data->ref_channel == 0)
		  data->pending.push(frame);
	       else
		  process_test_burst(frame, data);
	       
	       ++data->num_left;
	    }
	    ++input;
	    
	    // right (essentially the same)
	    
	    if ((data->num_right == 0
		 || (frame - data->last_burst_right) > SKIP_MARGIN
		 * data->msg_duration)
		&& (*input > THRESHOLD || *input < -THRESHOLD))
	    {
	       data->last_burst_right = frame;
	       // if this was the first burst, it's the reference channel
	       if (data->ref_channel == -1)
		  data->ref_channel = 1;
	       
	       if (data->ref_channel == 1)
		  data->pending.push(frame);
	       else
		  process_test_burst(frame, data);
	       
	       ++data->num_right;
	    }
	    ++input;
	 }
	 // multi-msg burst processing code
	 else
	 {
	    // left
	    
	    // If we're looking for a burst start, anything over threshold
	    // signals it.
	    if (!data->in_burst_left && (*input > THRESHOLD || *input < -THRESHOLD))
	    {
	       data->last_burst_left = frame;
	       data->end_candidate_left = frame;
	       // if this was the first burst, it's the reference channel
	       if (data->ref_channel == -1)
		  data->ref_channel = 0;
	       
	       if (data->ref_channel == 0)
		  data->pending.push(frame);
	       else
		  process_test_burst(frame, data);
	       
	       ++data->num_left;
	       data->in_burst_left = true;
	    }
	    // If we're looking for an end, need SUSTAIN_FACTOR
	    // times the expected duration of sub-threshold values to count it,
	    // otherwise we assume it was just a gap within the same burst.
	    else if (data->in_burst_left)
	    {
	       if (*input > THRESHOLD || *input < -THRESHOLD)
		  data->end_candidate_left = frame;
	       else if (frame - data->end_candidate_left
			> SUSTAIN_FACTOR * data->msgs * data->msg_duration)
	       {
		  // found the end, so add to the width histogram
		  update_width_histogram(data->last_burst_left, data->end_candidate_left, (data->ref_channel == 0) ? 0 : 1, data);

		  data->in_burst_left = false;
	       }
	    }
	    ++input;
	    
	    // right (essentially the same)
	    
	    if (!data->in_burst_right && (*input > THRESHOLD || *input < -THRESHOLD))
	    {
	       data->last_burst_right = frame;
	       data->end_candidate_right = frame;
	       // if this was the first burst, it's the reference channel
	       if (data->ref_channel == -1)
		  data->ref_channel = 1;
	       
	       if (data->ref_channel == 1)
		  data->pending.push(frame);
	       else
		  process_test_burst(frame, data);
	       
	       ++data->num_right;
	       data->in_burst_right = true;
	    }
	    else if (data->in_burst_right)
	    {
	       if (*input > THRESHOLD || *input < -THRESHOLD)
		  data->end_candidate_right = frame;
	       else if (frame - data->end_candidate_right
			> SUSTAIN_FACTOR * data->msgs * data->msg_duration)
	       {
		  // found the end, so add to the width histogram
		  update_width_histogram(data->last_burst_right, data->end_candidate_right, (data->ref_channel == 1) ? 0 : 1, data);

		  data->in_burst_right = false;
	       }
	    }
	    ++input;
	 }
      }
   }

   data->frames += framesPerBuffer;

   return 0;
}

// process a burst onset in the test stream
void process_test_burst(unsigned int frame, AnalysisData* data)
{
   // ref should always come before the corresponding test (abort if not)
   if (data->pending.empty())
   {
      data->cancel = true;
      cerr << "*** Aborting: found a test msg before corresponding ref" << endl
	   << "    (Possibly a highly-stretched burst)" << endl;
      return;
   }

   const int latency = frames_to_tenths_of_ms(frame - data->pending.front());

   // add current latency to circular buffer if at or above threshold
   if (SAVE_LATENCIES)
   {
      if (latency >= LATENCY_SAVE_THRESHOLD
	  && data->saved_latencies_num < LATENCIES_TO_SAVE)
      {
	 data->saved_latency_values[data->saved_latencies_num] = latency;
	 data->saved_latency_times[data->saved_latencies_num]
	    = frames_to_tenths_of_ms(frame);
	 ++data->saved_latencies_num;
      }
   }

   // If latency is greater than period, things are bunching up, so signal
   // ..that we should abort.
   if (latency > data->period)
   {
      data->cancel = true;
      cerr << "*** Aborting: found a latency greater than the period" << endl
	   << "    (This causes messages to bunch up and desynchronize)"
	   << endl;
      return;
   }

   data->pending.pop();

   if (latency < PA_LATENCY_HISTOGRAM_RANGE)
      ++(data->latency[latency / PA_LATENCY_HISTOGRAM_RESOLUTION]);
   else
      ++(data->out_of_range_latency);
   
   if (latency > data->max_latency)
      data->max_latency = latency;

   // only calculate jitter if it's not the first event pair
   if (data->num_left > 0 && data->num_right > 0)
   {
      // jitter in tenths of a millisecond
      const int jitter = latency - data->last_latency;
      
      if (jitter < PA_JITTER_HISTOGRAM_RANGE)
	 ++(data->jitter[jitter / PA_JITTER_HISTOGRAM_RESOLUTION]);
      else
	 ++(data->out_of_range_jitter);
      
      if (jitter > data->max_jitter)
	 data->max_jitter = jitter;
   }
   data->last_latency = latency;
   data->sum += latency;
   data->sumSq += latency*latency;
}

void update_width_histogram(int start, int end, int ref_test, AnalysisData* data)
{
   int width = frames_to_tenths_of_ms(end - start);

   // If width is greater than period, things are bunching up, so signal
   // ..that we should abort.
   if (width > data->period)
   {
      data->cancel = true;
      return;
   }

   if (ref_test == 0)
   {
      if (width < PA_WIDTH_HISTOGRAM_RANGE)
	 ++(data->ref_width[width / PA_WIDTH_HISTOGRAM_RESOLUTION]);
      else
	 ++(data->out_of_range_ref_width);
      
      if (width > data->max_ref_width)
	 data->max_ref_width = width;

      data->sum_ref_width += width;
      data->sumSq_ref_width += width*width;
   }
   else
   {
      if (width < PA_WIDTH_HISTOGRAM_RANGE)
	 ++(data->test_width[width / PA_WIDTH_HISTOGRAM_RESOLUTION]);
      else
	 ++(data->out_of_range_test_width);
      
      if (width > data->max_test_width)
	 data->max_test_width = width;

      data->sum_test_width += width;
      data->sumSq_test_width += width*width;
   }
}

unsigned int frames_to_tenths_of_ms(unsigned int frames)
{
   // time in tenths of an ms = frames * 10000 / SAMPLE_RATE
   // Clever formula to avoid rounding/overflow: a*b/c = (a/c)*b + (a%c)*b/c
   return (frames / SAMPLE_RATE) * 10000
      + (frames % SAMPLE_RATE) * 10000 / SAMPLE_RATE;
}


void Pa_Debug(PaError err)
{
   if (err != paNoError)
   {
      cerr << "PortAudio error: " << Pa_GetErrorText(err) << endl;
      exit(-1);
   }
}
