/*
 * PortAudio analysis portion of the MIDI testing module.
 *
 * See test.cpp for details.
 */

#ifndef PA_ANALYSIS_HPP
#define PA_ANALYSIS_HPP

#include <portaudio.h>
#include <iostream>
#include <queue>
#include <string>

using namespace std;

const int SAMPLE_RATE = 44100;
const float THRESHOLD = 0.3;            // Value over (or neg. under) is "on"
const float SKIP_MARGIN = 1.8;		// Skip over this many times the
					// ..expected transmission time, in
					// ..the single-msg case.
const float SUSTAIN_FACTOR = 2.5;	// In the multi-msg case, require
 					// ..this many times the expected
					// ..burst width of sub-threshold
					// ..values before we decide it's over.
const int PERIODS_TO_SAVE = 5;		// Periods worth of data to dump to
					// ..a datafile if we abort.
const bool SAVE_LATENCIES = true;	// Save latencies over a threshold
					// ..to a file?
					// ..(Useful for temporal analysis)
const int LATENCY_SAVE_THRESHOLD = 70;  // Threshold in tenths of a ms at
                                        // ..which to save latency readings
                                        // ..to log to a datafile.
					// ..Note: this does not affect the
					// ..histograms, which include all
					// ..latencies.
const int LATENCIES_TO_SAVE = 5000;     // Number of latency readings
                                        // ..above the threshold to
                                        // ..write out to the datafile.
const string AUDIO_DEBUG_FILENAME = "debug.raw";
const string LATENCIES_FILENAME   = "latencies";

// to make things easy, range should be an integer multiple of resolution
// for each histogram
const int PA_LATENCY_HISTOGRAM_RESOLUTION = 2;	// in tenths of a ms
const int PA_LATENCY_HISTOGRAM_RANGE = 200;     // range is up to this (x0.1ms)
const int PA_LATENCY_HISTOGRAM_BINS =
   PA_LATENCY_HISTOGRAM_RANGE / PA_LATENCY_HISTOGRAM_RESOLUTION;

const int PA_JITTER_HISTOGRAM_RESOLUTION = 2;
const int PA_JITTER_HISTOGRAM_RANGE = 100;
const int PA_JITTER_HISTOGRAM_BINS =
   PA_JITTER_HISTOGRAM_RANGE / PA_JITTER_HISTOGRAM_RESOLUTION;

const int PA_WIDTH_HISTOGRAM_RESOLUTION = 2;
const int PA_WIDTH_HISTOGRAM_RANGE = 2000;
const int PA_WIDTH_HISTOGRAM_BINS =
   PA_WIDTH_HISTOGRAM_RANGE / PA_WIDTH_HISTOGRAM_RESOLUTION;


// data collected by the callback
struct AnalysisData
{
   // first something it needs to know from main()
   long msg_duration;			// estimated message duration in frames
   int msgs;				// messages per burst
   int period;				// period in units of 0.1ms

   // signal that there was a glitch (latency too high) so we should cancel
   bool cancel;
   // stuff to help us debug when this glitch happens
   int period_in_frames;
   float* saved_samples; 
   int saved_samples_index;

   // save first LATENCIES_TO_SAVE latencies for temporal analysis
   int saved_latencies_num;
   int saved_latency_values[LATENCIES_TO_SAVE];
   unsigned int saved_latency_times[LATENCIES_TO_SAVE];

   // 0 if reference is left, 1 if reference is right, -1 if not detected yet
   int ref_channel;

   int latency[PA_LATENCY_HISTOGRAM_BINS];
   int jitter[PA_JITTER_HISTOGRAM_BINS];
   int ref_width[PA_WIDTH_HISTOGRAM_BINS];
   int test_width[PA_WIDTH_HISTOGRAM_BINS];

   int out_of_range_latency;
   int out_of_range_jitter;
   int out_of_range_ref_width;
   int out_of_range_test_width;

   // last latency, for use in calculating jitter by the L_i - L_(i-1) method
   // (in tenths of a millisecond)
   int last_latency; 			// Not initialized (set dynamically)

   // max latency and jitter, in tenths of a ms
   int max_latency;
   int max_jitter;
   int max_ref_width;
   int max_test_width;

   long sum, sumSq;			// sum and sum of squares of latencies
   long sum_ref_width, sum_test_width, 
	    sumSq_ref_width, sumSq_test_width;

   // number of bursts we've seen in each channel so far
   int num_left, num_right;

   // frames that have passed since we started
   unsigned int frames;

   // pending reference events not yet matched up to a test event
   // (without some serious problems the reverse should never happen)
   // the values recorded are the time measured in frames
   queue<int> pending;

   // time (in frames) the last burst was recorded in each channel
   int last_burst_left, last_burst_right;

   // currently within a burst, looking for its end (otherwise looking for
   // a burst start)
   bool in_burst_left, in_burst_right;

   // candidate for a burst end (which will pan out of sub-threshold sustains
   // for a sufficient time)
   // time is in frames
   int end_candidate_left, end_candidate_right;
   					// Not initialized (set dynamically)
};

void initialize_analysis_data(AnalysisData& data);
void initialize_analysis_data_saved_samples(AnalysisData& data);
void writeout_saved_samples_data(AnalysisData& data, ostream& file);
void writeout_saved_latencies_data(AnalysisData& data, ostream& file);
void Pa_Debug(PaError err);
int analysis_callback(void *inputBuffer, void *outputBuffer,
		      unsigned long framesPerBuffer,
		      PaTimestamp outTime, void *userData);
void process_test_burst(unsigned int frame, AnalysisData* data);
void update_width_histogram(int start, int end, int ref_test, AnalysisData* data);
unsigned int frames_to_tenths_of_ms(unsigned int frames);

#endif
