#include "pm_callback.hpp"

extern MidiData midiData;

using namespace std;

void initialize_midi_data(MidiData& data)
{
   for (int i = 0; i < PM_LATENCY_HISTOGRAM_BINS; ++i)
   {
      data.histogram[i] = 0;
      data.histogram_neg[i] = 0;
      data.histogram_out[i] = 0;
      data.histogram_in[i] = 0;
   }

   data.num_iterations = 0;

   data.sum = data.sumSq = data.sum_out = data.sumSq_out = data.sum_in = data.sumSq_in = 0;

   data.max_latency = 0;
   data.max_neg_latency = 0;
   data.max_latency_out = 0;
   data.max_latency_in = 0;
   data.out_of_range = 0;
   data.out_of_range_neg = 0;
   data.out_of_range_out = 0;
   data.out_of_range_in = 0;

   data.previous_callback_time = 0;

   data.settled = false;
   data.test_in = false;
   data.test_out = false;
   data.test_thru = false;

   data.note_on = false;
}

PtTimestamp Pt_Time_wrapper(void* dummy)
{
   (void)dummy;
   return Pt_Time();
}

bool Pt_is_settled(MidiData * data)
{
   return data->settled;
}

/* PortTime callback: outputs and/or reads MIDI, and generates software
 * histograms */
void pt_callback(PtTimestamp timestamp, void *userData)
{
   MidiData* data = (MidiData*)userData;

   long difference = timestamp - data->previous_callback_time - data->period;
   data->previous_callback_time = timestamp;

   // internal API latency
   long latency_out = 0, latency_in = 0;
   
   // allow settleTime msec for the system to settle down
   if (timestamp < SETTLE_TIME)
      return;
   else if (!data->settled)
      // allow main program to know when this has occured...
      data->settled = true;

   ++data->num_iterations;

   // send note on/offs if so desired
   if (data->test_out)
   {
      PmEvent buffer[MAX_NOTES_PER_ITERATION];
      const PmTimestamp now = Pt_Time();
      
      for (int i = 0; i < data->out_num; ++i)
      {
	 buffer[i].timestamp = now;
	 if (data->out_type == MidiData::NOTES)
	 {
	    // note on if notes are off and vice-versa
	    if (data->note_on)
	       buffer[i].message = Pm_Message(0x90, i, 0);
	    else
	       buffer[i].message = Pm_Message(0x90, i, 100);
	 }
	 else // ACTIVESENSE
	 {
	    buffer[i].message = Pm_Message(0xFE, 0, 0);
	 }
      }
      if (data->out_type == MidiData::NOTES)
	 data->note_on = !data->note_on;

      PmTimestamp startAPI = Pt_Time();
      Pm_DebugStream(Pm_Write(data->out, buffer, data->out_num), data->out);
      latency_out += Pt_Time() - startAPI;
   }

   // read all waiting events (if user requested)
   if (data->test_in)
   {
      PmError status;
      PmEvent buffer[1];

      PmTimestamp startAPI = Pt_Time();	// polling counts as "input" delay
      Pm_DebugStream((status = Pm_Poll(data->in)), data->in);
      while (status == TRUE)
      {
	 Pm_DebugStream(Pm_Read(data->in, buffer, 1), data->in);

	 Pm_DebugStream((status = Pm_Poll(data->in)), data->in);
      }
      latency_in += Pt_Time() - startAPI;
   }

   // pass midi through (if desired)
   if (data->test_thru)
   {
      PmError status;
      PmEvent buffer[1];

      PmTimestamp startAPI = Pt_Time();
      Pm_DebugStream((status = Pm_Poll(data->in)), data->in);
      latency_in += Pt_Time() - startAPI;
      while (status == TRUE)
      {
	 startAPI = Pt_Time();
	 Pm_DebugStream(Pm_Read(data->in, buffer, 1), data->in);
	 latency_in += Pt_Time() - startAPI;

	 startAPI = Pt_Time();
	 Pm_DebugStream(Pm_Write(data->out, buffer, 1), data->out);
	 latency_out += Pt_Time() - startAPI;

 	 startAPI = Pt_Time();
	 Pm_DebugStream((status = Pm_Poll(data->in)), data->in);
	 latency_in += Pt_Time() - startAPI;
      }
   }

   // update the histograms
   if (difference < 0)
   {
      if (-difference < PM_LATENCY_HISTOGRAM_BINS)
	 ++data->histogram_neg[-difference];
      else
	 ++data->out_of_range_neg;
   }
   else if (difference < PM_LATENCY_HISTOGRAM_BINS)
      ++data->histogram[difference];
   else
      ++data->out_of_range;

   if (latency_out < PM_LATENCY_HISTOGRAM_BINS)
      ++data->histogram_out[latency_out];
   else
      ++data->out_of_range_out;
   
   if (latency_in < PM_LATENCY_HISTOGRAM_BINS)
      ++data->histogram_in[latency_in];
   else
      ++data->out_of_range_in;
   
   if (data->max_latency < difference)
      data->max_latency = difference;
   if (difference < data->max_neg_latency)
      data->max_neg_latency = difference;

   if (data->max_latency_out < latency_out)
      data->max_latency_out = latency_out;
   if (data->max_latency_in < latency_in)
      data->max_latency_in = latency_in;

   data->sum += difference;
   data->sumSq += difference*difference;
   data->sum_out += latency_out;
   data->sumSq_out += latency_out*latency_out;
   data->sum_in += latency_in;
   data->sumSq_in += latency_in*latency_in;
}

#include<fstream>
// The following can be used to store current state to file;
// useful for Perl scripts that help automate testing.
void Pm_DisplayInputDevice(std::ostream &to)
{
   /* list device information */
   for (int i = 0; i < Pm_CountDevices(); ++i)
   {
      const PmDeviceInfo *info = Pm_GetDeviceInfo(i);
      if (info->input)
      {
	 to << i << " : " <<  info->interf << ", " <<  info->name << endl;
      }
   }
}
int Pm_PromptForInputDevice() 
{
   cerr << "Choose MIDI input device:" << endl;
   Pm_DisplayInputDevice(std::cerr);
   cerr << "? ";
   int input;
   cin >> input;
   return input;
}
void Pm_DisplayOutputDevice(std::ostream &to)
{
   /* list device information */
   for (int i = 0; i < Pm_CountDevices(); ++i)
   {
      const PmDeviceInfo *info = Pm_GetDeviceInfo(i);
      if (info->output)
      {
	 to << i << " : " <<  info->interf << ", " <<  info->name << endl;
      }
   }
}
int Pm_PromptForOutputDevice()
{
   cerr << "Choose MIDI output device:" << endl;
   Pm_DisplayOutputDevice(std::cerr);
   cerr << "? ";
   int input;
   cin >> input;
   return input;
}

void Pm_Debug(PmError error)
{
   /* note that errors are negative and some routines return
    * positive values to indicate success status rather than error
    */
   if (error < 0)
   {
      cerr << "PortMidi call failed... "
	   << Pm_GetErrorText(error) << endl;
      exit(-1);
   }
}

void Pm_DebugStream(PmError error, PortMidiStream* stream)
{
   if (error == pmHostError)
   {
      char msg[PM_HOST_ERROR_MSG_LEN];
      cerr << "HostError: ";
      /* this function handles bogus stream pointer */
      Pm_GetHostErrorText(stream, msg, PM_HOST_ERROR_MSG_LEN);
      cerr << msg << endl;
      exit(-1);
   }
   else if (error < 0)
   {
      Pm_Debug(error);
   }
}
