/*
 * MIDI Latency testing module.
 *
 * Runs tests of MIDI latency, as the name implies.  Sends MIDI test
 * messages from a reference computer, route them through on a test
 * computer, and measures the round-trip time.  See README for
 * instructions and details.
 *
 * Authors: Mark Nelson <mnelson@hmc.edu> and Belinda Thom <bthom@cs.hmc.edu>
 *
 * Incorporates some code originally from 'latency.c' from the
 *   PortMidi distribution. <http://www.cs.cmu.edu/~music/portmusic/>
 *
 * Written summer/fall 2003; last modified May 22, 2004.
 */

#include <iostream>
#include <fstream>
#include <iomanip>
#include <stdlib.h>
#include <assert.h>
#undef NDEBUG // Don't let VC override the assertions in Release mode
#include <ctype.h>
#include <math.h>
#ifndef WIN32
#include <unistd.h>
#endif
#ifdef WIN32
#include <windows.h>
#endif
#include <portaudio.h>
#include <portmidi.h>
#include <porttime.h>

#include "pm_callback.hpp"
#include "pa_analysis.hpp"

const float PERIOD_FACTOR = 2.0;	// Require period to be this many
					// ..times longer than expected
					// ..burst length (plus the sustain
					// ..duration if >1-note bursts).

#ifndef BIG_MEM
const int LOAD_MATRIX_SIZE = 289;	// Makes about 1M load
					// ..(double + int = 12 bytes per element)
#else
const int LOAD_MATRIX_SIZE = 5000;	// Load test will use two matrices,
					// ..each this large, square.
#endif

const int RAND_SEED = 12345;		// To allow test repeatability.
					// ..(Used in the load-generation).

// global because of difficulty allocating large local arrays statically
double dbl_matrix[LOAD_MATRIX_SIZE][LOAD_MATRIX_SIZE];
int int_matrix[LOAD_MATRIX_SIZE][LOAD_MATRIX_SIZE];

using namespace std;

void initialize_load();
void generate_load(int duration);

#ifndef WIN32
int min(int a, int b) { return a < b ? a : b; }
int max(int a, int b) { return a > b ? a : b; }
#endif

#include<fstream>

// General comment:
//   Rationale for cerr/cout usage:
// 	 - stuff sent to cerr is for interaction w/user.
//	 - stuff sent to cout is what we want to save as a report log.
//

int main()
{
   int duration;
   bool software_stats, save_debug;
   bool load;
   enum { HARDWARE, SOFTWARE, MIDITHRU, ASKIN, ASKOUT } testType;
   int n;
   int i;

   MidiData midiData;
   initialize_midi_data(midiData);
   AnalysisData analysisData;
   initialize_analysis_data(analysisData);
   PortAudioStream* input;
	
   // Query user for type of test to run
   cerr << "Would you like to do (see README for descriptions):" << endl
	<< "\t1. A hardware test" << endl
	<< "\t2. A software test" << endl
	<< "\t3. MIDI-thru" << endl
	<< "\t4. Display MIDI in devices" << endl
	<< "\t5. Display MIDI out devices" << endl
	<< endl;
   cin >> n;
   if (n == 1)
      // Ref machine pulses messages to Test machine and records
      // ..transcoder audio
      testType = HARDWARE;
   else if (n == 2)
      // This test exercises periodicity of callback timer
      // ..under various configs (Midi in, out, none) that are set later
      testType = SOFTWARE;
   else if (n == 3)
      // Test machine passes messages thru to transcoder
      testType = MIDITHRU;
   else if (n == 4)
      // This test added to allow test machine to querey its midi 
      // devices. This allows outer perl script to assign PM devices
      // (numbers) using smarter string matching (on device names).
      testType = ASKIN;
   else if (n == 5)
      // ditto
      testType = ASKOUT;
   else
      assert(0);

   Pm_Debug(Pm_Initialize()); // Might as well do this explicitly
   
   if (testType == ASKIN) 
   {
      // This data stored to stdout so perl script with saving STDOUT works 
      cout << "Midi input devices:" << endl;
      Pm_DisplayInputDevice(cout);
      cout.flush();
      return 0;
	
   } else if (testType == ASKOUT) 
   {
      // Ditto
      cout << "Midi output devices:" << endl;
      Pm_DisplayOutputDevice(cout);
      cout.flush();
      return 0;

   } else if (testType == HARDWARE)
   {
      cerr << "Also output software stats for the current machine (y/n)? ";
      char yn;
      cin >> yn;
      if (toupper(yn) == 'Y')
	 software_stats = true;
      else if (toupper(yn) == 'N')
	 software_stats = false;
      else
	 assert(0);

      cerr << "Save recent audio data when test fails (y/n)? ";
      cin >> yn;
      if (toupper(yn) == 'Y')
	 save_debug = true;
      else if (toupper(yn) == 'N')
	 save_debug = false;
      else
	 assert(0);

      midiData.test_in = false;
      midiData.test_out = true;

   } else if (testType == SOFTWARE) 
   {
      software_stats = true;
      cerr << "What sort of software test? " << endl
	   << "\t1. Scheduler only (no MIDI)" << endl
	   << "\t2. Read MIDI input" << endl
	   << "\t3. Write MIDI output" << endl
	   << "\t4. Pass MIDI thru" << endl
	   << "? ";
      int softwareType;
      cin >> softwareType;
      switch (softwareType) {
      case 1: 
	 // scheduler only: nothing to turn on
	 break;
      case 2:
	 midiData.test_in = true;
	 break;
      case 3:
	 midiData.test_out = true;
	 break;
      case 4:
	 midiData.test_thru = true;
	 break;
      default: 
	 assert(0);
      }	
      cerr << "Choose callback timer period (in ms): ";
      cin >> midiData.period;
      assert(midiData.period >= 1);

   } else if (testType == MIDITHRU)
   {
      software_stats = true;
      midiData.test_thru = true;
      midiData.period = 1;		// As responive as possible!

   } else				// Unknown test type
      assert(0);

   // Query user whether or not to test functionality under load
   if (testType != HARDWARE)
   {
      cerr << "Simulate CPU/memory load (y/n)? ";
      char yn;
      cin >> yn;
      if (toupper(yn) == 'Y')
	 load = true;
      else if (toupper(yn) == 'N')
	 load = false;
      else
	 assert(0);
   }
   else					// No load option for HARDWARE
   {
      load = false;
   }

   // Query user how long to loop
   cerr << "Choose duration of test (in s): ";
   cin >> duration;
   assert(duration >= 1);
      
   if (midiData.test_in || midiData.test_thru)
   {
      // Open portmidi input stream
      int i = Pm_PromptForInputDevice();
      Pm_DebugStream(Pm_OpenInput(&midiData.in, 
				  i,
				  NULL, 
				  INPUT_BUFFER_SIZE, 
				  Pt_Time_wrapper,
				  &midiData, 
				  NULL), // no thru (do it manually)
		     NULL);	        // This last param required but
					// ..not used for debugging 
		
      // don't filter active sensing messages 
      // ..(or that test won't work...)
      Pm_SetFilter(midiData.in, 0);

   } 
   
   if (midiData.test_out || midiData.test_thru)
   {
      // Open portmidi output stream
      int i = Pm_PromptForOutputDevice();
      Pm_DebugStream(Pm_OpenOutput(&midiData.out, 
				   i,
				   NULL,
				   OUTPUT_BUFFER_SIZE,
				   Pt_Time_wrapper,
				   &midiData, 
				   0),	// don't do latency 
					// ..(i.e. Midi Stream) scheduling 
		     NULL);		// for debugging
      
      // Note: SOFTWARE/MIDITHRU period already determined
      if (testType == HARDWARE) 
      {
	 cerr << "Output how often (in ms)? ";
	 cin >> midiData.period;
	 assert(midiData.period >= 1);
	 analysisData.period = midiData.period * 10;
	 initialize_analysis_data_saved_samples(analysisData);
      }
   }
   
   if (midiData.test_out)
   {
      cerr << "How many messages at a time? ";
      int msgs;
      cin >> msgs;
      midiData.out_num = msgs;
      analysisData.msgs = msgs;
      assert(msgs >= 1);
      assert(msgs <= MAX_NOTES_PER_ITERATION);
      cerr << endl;
      cerr << "Choose message type:" << endl
	   << "\t1. Note-on and note-offs" << endl
	   << "\t2. Active sensing (0xFE) bytes" << endl;
      int n;
      cin >> n;
      if (n == 1)
      {
	 midiData.out_type = MidiData::NOTES;
	 analysisData.msg_duration = SAMPLE_RATE / 1000; // 1 ms
      }
      else if (n == 2)
      {
	 midiData.out_type = MidiData::ACTIVESENSE;
	 analysisData.msg_duration = SAMPLE_RATE / 3000; // 1/3 ms
      }
      else
      {
	 assert(0);
      }
   }   
   cerr << endl;
	
   // Make sure we have enough margin between period and burst duration
   if (testType == HARDWARE)
   {
      assert((SAMPLE_RATE / 1000 * midiData.period)
	     > (PERIOD_FACTOR * (1.0 + SUSTAIN_FACTOR)
		* analysisData.msg_duration * analysisData.msgs));
   }

   if (load)
   {
      cerr << "Initializing load matrix (size " << LOAD_MATRIX_SIZE << ")..." << endl;
      initialize_load();
   }

   // Query user to start (allows manual synch up)
   if (testType == HARDWARE)
   {
      cerr << "Enter any key to start REF machine " << duration  + (SETTLE_TIME/1000.) << " [sec] hardware test" << endl;
      cerr << "(wait for other TEST machine to settle first)." << endl;
   } else
   {
      cerr << "Enter any key to start TEST machine " << duration + (SETTLE_TIME/1000.) << " [sec] " 
	   << ((testType == MIDITHRU) ? " midi-thru " : " software ") << "test" << endl;

      cerr << "(this machine should start before REF and have longer duration)." << endl;
   }
   char c;
   cin >> c;

   // Start one or both PortMidi callbacks
   Pt_Start(midiData.period, &pt_callback, &midiData);

   if (testType == HARDWARE) 
   {
      // Enable audio recording
      Pa_Debug(Pa_Initialize());
      Pa_Debug(Pa_OpenDefaultStream(&input, 
				    2,          // stereo input
				    0,	        // no output 
				    paFloat32,	// 32 bit floating point input 
				    SAMPLE_RATE,
				    1024,	// frames per buffer 
				    0,		// number of buffers; zero 
					        // ..means the default minimum 
				    analysis_callback, // callback function 
				    &analysisData));   // data to pass to callback 
      Pa_Debug(Pa_StartStream(input));
   }

   cerr << "Test has begun running..." << endl; 
   PtTimestamp start = Pt_Time();

   int loopInterval = 1; // approximately how many seconds each loop takes
   bool oneShot = true;
   // Include settle time in explicit wait loop
   int totalDuration = SETTLE_TIME + (duration * 1000);
   // 5 minute or second clock updates depending on test duration.
   int updateInterval = (int)((totalDuration > 120000) ? (5*60.)/loopInterval : (5./loopInterval));
   PtTimestamp diff;
   unsigned int ct = 0;
   while ((diff = Pt_Time() - start) < totalDuration)
   {
      if (midiData.settled && oneShot)
      {
	 oneShot = false;
	 cerr << "\tMachine has settled, now processing MIDI..." << endl;
      }

      // Update occassionally how much is left
      // ..(very useful when running long tests)...
      if ((ct++ % updateInterval) == 0)
      {
	 int remain = (int)((totalDuration - diff)/1000.0);
	 cerr << "\tTest has approx " << remain << " seconds left... " << endl;
      }

      if ((testType == HARDWARE) && analysisData.cancel)
	 break;
	
      if ((testType != HARDWARE) && load)
      {
	 generate_load(loopInterval);	
      } else
      {
	 // The standard do nothing...
#ifdef WIN32
	 Sleep(loopInterval*1000);
#else
	 sleep(loopInterval);
#endif
      }
   }
   PmTimestamp stop = Pt_Time();
   // Stop MIDI processing
   Pt_Stop();

   if (testType == MIDITHRU)
   {
      cerr << "Test finished!" << endl;
   }
   else if (testType == HARDWARE)
   {
      // Wait the last little bit audio from TEST machine 
      // ..might take a bit to get processed.
#ifdef WIN32
      Sleep(STRAGGLE_TIME);
#else
      sleep(STRAGGLE_TIME/1000);
#endif
      Pa_Debug(Pa_StopStream(input));
      Pa_Debug(Pa_CloseStream(input));
      Pa_Debug(Pa_Terminate());
   }

   Pm_Debug(Pm_Terminate());		// DebugStream not needed 
					// ..(Pm_Terminate prints errors)

   // Output test configuration
   cout << "Test type: ";
   if (testType == HARDWARE)
      cout << "hardware";
   else if (testType == SOFTWARE)
      cout << "software";
   else
   { 
      assert(testType == MIDITHRU);
      cout << "thru";
   }
   cout << endl;
   cout << "Duration: " << max(0, stop - start) * 0.001 << " seconds" << endl;
   cout << "Callback period: " << midiData.period << " ms" << endl;
   if (midiData.test_out)
      cout << "Messages per callback: " << midiData.out_num << endl;
   if (testType != HARDWARE)
      cout << "Load: " << (load ? "yes" : "no") << endl;
   if (load)
   {
      cout << "Load matrix size: " << LOAD_MATRIX_SIZE << endl;
      cout << "Random seed: " << RAND_SEED << endl;
   }
   if (testType == HARDWARE)
   {
      cout << "Sample rate: " << SAMPLE_RATE << " Hz" << endl;
      cout << "Threshold value: " << THRESHOLD << endl;
      if (analysisData.msgs == 1)
	 cout << "Skip margin (single-msg): " << SKIP_MARGIN << endl;
      else
	 cout << "Sustain factor (multi-msg): " << SUSTAIN_FACTOR << endl;
   }
   cout << endl;

   // Output measured data
   cout << "Note: all histogram bin labels are the left (towards zero) edge" << endl
	<< "      of the bin, except between-callbacks software latency, where"	<< endl
	<< "      they are the middle of the bin (since it goes - and +)." << endl << endl;

   if (analysisData.cancel && save_debug)
   {
      cerr << "** WARNING: Test aborted!" << endl << endl;

      cout << "** WARNING: Test aborted!" << endl << endl;

      // Write dump of the last few periods to a datafile
      // ..(raw floating point format, interleaved stereo samples)
      ofstream outfile(AUDIO_DEBUG_FILENAME.c_str());
      writeout_saved_samples_data(analysisData,outfile);
      outfile.close();
   }

   if (testType == HARDWARE)
   {
      // Check to see if any ref messages went un-matched;
      // ..(can't be the other way around or analysis code would've asserted)
      if (analysisData.num_left != analysisData.num_right)
	 cout << "** WARNING: Did not match up " 
	      << (analysisData.ref_channel == 0 ? 
		  analysisData.num_left : analysisData.num_right)
	      << " reference messages with"
	      << endl
	      << "**          corresponding " 
	      << (analysisData.ref_channel == 0 ? 
		  analysisData.num_right : analysisData.num_left)
	      << " test messages!" << endl << endl;
      // if multi-msg burst test, make sure there's no open bursts
      if (analysisData.msgs > 1
	  && (analysisData.in_burst_left || analysisData.in_burst_right))
      {
	 cout << "** WARNING: Burst still open at the end of analysis."
	      << endl << endl;
      }
      // Output PortAudio analysis stats
      cout << "Round-trip (hardware) latency histogram:" << endl;
      cout << "Total number of counts is " << analysisData.num_left << endl;
      cout << "Times are in 10^-4 s (tenths of a millisecond)" << endl;
      cout << "Latency      Number of occurrences" << endl;
		
      int len = min(PA_LATENCY_HISTOGRAM_BINS,
		    (analysisData.max_latency / PA_LATENCY_HISTOGRAM_RESOLUTION) + 1);
      for (i = 0; i < len; ++i)
	 cout << setw(2) << (i * PA_LATENCY_HISTOGRAM_RESOLUTION)
	      << "      " << setw(10)
	      << analysisData.latency[i] << endl;
		
      cout << "Number of points above range: "
	   << analysisData.out_of_range_latency << endl;
      cout << "Maximum latency: " << analysisData.max_latency << endl;
      cout << "Mean latency: " << ((double)analysisData.sum / analysisData.num_left) << endl;
      cout << "Stddev latency: "
	   << sqrt(((double)analysisData.sumSq
		    - ((double)analysisData.sum
		       * ((double)analysisData.sum / analysisData.num_left)))
		   / ((double)analysisData.num_left - 1))
	   << endl;
      cout << endl;
		
      cout << "Round-trip (hardware) jitter histogram:" << endl;
      cout << "Times are in 10^-4 s (tenths of a millisecond)" << endl;
      cout << "Latency      Number of occurrences" << endl;
		
      len = min(PA_JITTER_HISTOGRAM_BINS,
		(analysisData.max_jitter / PA_JITTER_HISTOGRAM_RESOLUTION) + 1);
      for (i = 0; i < len; ++i)
	 cout << setw(2) << (i * PA_JITTER_HISTOGRAM_RESOLUTION)
	      << "      " << setw(10)
	      << analysisData.jitter[i] << endl;
		
      cout << "Number of points above range: "
	   << analysisData.out_of_range_jitter << endl;
      cout << "Maximum jitter: " << analysisData.max_jitter << endl;
      cout << endl;
		
      // output burst width histograms
      if (analysisData.msgs > 1)
      {
	 cout << "REF burst width histogram:" << endl;
	 cout << "Times are in 10^-4 s (tenths of a millisecond)" << endl;
	 cout << "Width        Number of occurrences" << endl;
			
	 int len = min(PA_WIDTH_HISTOGRAM_BINS,
		       (analysisData.max_ref_width / PA_WIDTH_HISTOGRAM_RESOLUTION) + 1);
	 for (i = 0; i < len; ++i)
	    cout << setw(2) << (i * PA_WIDTH_HISTOGRAM_RESOLUTION)
		 << "      " << setw(10)
		 << analysisData.ref_width[i] << endl;

	 cout << "Number of points above range: "
	      << analysisData.out_of_range_ref_width << endl;
	 cout << "Maximum width: " << analysisData.max_ref_width << endl;
	 cout << "Mean width: " << ((double)analysisData.sum_ref_width / analysisData.num_left) << endl;
	 cout << "Stddev width: "
	      << sqrt(((double)analysisData.sumSq_ref_width
		       - ((double)analysisData.sum_ref_width
			  * ((double)analysisData.sum_ref_width / analysisData.num_left)))
		      / ((double)analysisData.num_left - 1))
	      << endl;
	 cout << endl;
			
	 cout << "TEST burst width histogram:" << endl;
	 cout << "Times are in 10^-4 s (tenths of a millisecond)" << endl;
	 cout << "Width        Number of occurrences" << endl;
			
	 len = min(PA_WIDTH_HISTOGRAM_BINS,
		   (analysisData.max_test_width / PA_WIDTH_HISTOGRAM_RESOLUTION) + 1);
	 for (i = 0; i < len; ++i)
	    cout << setw(2) << (i * PA_WIDTH_HISTOGRAM_RESOLUTION)
		 << "      " << setw(10)
		 << analysisData.test_width[i] << endl;
			
	 cout << "Number of points above range: "
	      << analysisData.out_of_range_test_width << endl;
	 cout << "Maximum width: " << analysisData.max_test_width << endl;
	 cout << "Mean width: " << ((double)analysisData.sum_test_width / analysisData.num_left) << endl;
	 cout << "Stddev width: "
	      << sqrt(((double)analysisData.sumSq_test_width
		       - ((double)analysisData.sum_test_width
			  * ((double)analysisData.sum_test_width / analysisData.num_left)))
		      / ((double)analysisData.num_left - 1))
	      << endl;
	 cout << endl;
      }

      // write out the saved latencies to a file, if requested
      // (one per line, plaintext, in tenths of a millisecond)
      if (SAVE_LATENCIES)
      {
	 ofstream outfile(LATENCIES_FILENAME.c_str());
	 writeout_saved_latencies_data(analysisData, outfile);
	 outfile.close();
      }
   }

   if (software_stats)
   {
      // Output software (portmidi) callback stats
      cout << "Software latency histogram: between callbacks" << endl;\
									 cout << "Times are in ms" << endl;
      cout << "Latency      Number of occurrences" << endl;
      int len = min(PM_LATENCY_HISTOGRAM_BINS, -midiData.max_neg_latency + 1);
      for (i = len - 1; i > 0; --i)
	 cout << setw(2) << -i << "      " << setw(10)
	      << midiData.histogram_neg[i] << endl;
		
      len = min(PM_LATENCY_HISTOGRAM_BINS, midiData.max_latency + 1);
      for (i = 0; i < len; ++i)
	 cout << setw(2) << i << "      " << setw(10) << midiData.histogram[i]
	      << endl;     
      cout << "Number of points below range: "
	   << midiData.out_of_range_neg << endl;
      cout << "Maximum negative latency: "
	   << midiData.max_neg_latency << endl;
      cout << "Number of points above range: "	   
	   << midiData.out_of_range << endl;
      cout << "Maximum latency: " << midiData.max_latency << endl;
      cout << "Mean latency: "
	   << ((double)midiData.sum / midiData.num_iterations) << endl;
      cout << "Stddev latency: "
	   << sqrt(((double)midiData.sumSq
		    - ((double)midiData.sum
		       * ((double)midiData.sum / midiData.num_iterations)))
		   / ((double)midiData.num_iterations - 1))
	   << endl;
      cout << endl;
		
      cout << "Software latency histogram: MIDI output API" << endl;
      cout << "Times are in ms" << endl;
      cout << "Latency      Number of occurrences" << endl;
      len = min(PM_LATENCY_HISTOGRAM_BINS, midiData.max_latency_out + 1);
      for (i = 0; i < len; ++i)
	 cout << setw(2) << i << "      " << setw(10)
	      << midiData.histogram_out[i] << endl;
      cout << "Number of points above range: "
	   << midiData.out_of_range_out << endl;
      cout << "Maximum latency: " << midiData.max_latency_out << endl;
      cout << "Mean latency: "
	   << ((double)midiData.sum_out / midiData.num_iterations) << endl;
      cout << "Stddev latency: "
	   << sqrt(((double)midiData.sumSq_out
		    - ((double)midiData.sum_out	       
		       * ((double)midiData.sum_out / midiData.num_iterations)))
		   / ((double)midiData.num_iterations - 1))
	   << endl;
      cout << endl;
		
      cout << "Software latency histogram: MIDI input API" << endl;
      cout << "Times are in ms" << endl;
      cout << "Latency      Number of occurrences" << endl;
      len = min(PM_LATENCY_HISTOGRAM_BINS, midiData.max_latency_in + 1);
      for (i = 0; i < len; ++i)
	 cout << setw(2) << i << "      " << setw(10)
	      << midiData.histogram_in[i] << endl;
      cout << "Number of points above range: "
	   << midiData.out_of_range_in << endl;
      cout << "Maximum latency: " << midiData.max_latency_in << endl;
      cout << "Mean latency: "
	   << ((double)midiData.sum_in / midiData.num_iterations) << endl;
      cout << "Stddev latency: "
	   << sqrt(((double)midiData.sumSq_in
		    - ((double)midiData.sum_in
		       * ((double)midiData.sum_in / midiData.num_iterations)))
		   / ((double)midiData.num_iterations - 1))
	   << endl;
   }

   cerr<< "Ready to exit" << endl;
	
   return 0;  // Exiting normally.
}

void initialize_load()
{
   const int N = LOAD_MATRIX_SIZE;	// for convenience

   assert(N < RAND_MAX);		// because of the way the code's written

   for (int i = 0; i < N; ++i)
   {
      for (int j = 0; j < N; ++j)
      {
	 dbl_matrix[i][j] = (double)rand() / RAND_MAX;
	 int_matrix[i][j] = rand();
      }
   }
}

void generate_load(int duration)
{
   const int N = LOAD_MATRIX_SIZE;	// for convenience

   PmTimestamp start = Pt_Time();

   srand(RAND_SEED);

   while (Pt_Time() < start + (duration * 1000))
   {
      for (int i = 0; i < 100; ++i)
      {
	 int a1 = rand() % N;
	 int b1 = rand() % N;
	 int a2 = rand() % N;
	 int b2 = rand() % N;
	 
	 switch (rand() % 8)
	 {
	 case 0: 
	    dbl_matrix[a1][b1] += dbl_matrix[a2][b2]; break;
	 case 1:
	    dbl_matrix[a1][b1] -= dbl_matrix[a2][b2]; break;
	 case 2:
	    dbl_matrix[a1][b1] *= dbl_matrix[a2][b2]; break;
	 case 3:
	    if (dbl_matrix[a2][b2] > 0.0001 || dbl_matrix[a2][b2] < -0.0001)
	       dbl_matrix[a1][b1] /= dbl_matrix[a2][b2];
	    break;

	 case 4:
	    int_matrix[a1][b1] += int_matrix[a2][b2]; break;
	 case 5:
	    int_matrix[a1][b1] -= int_matrix[a2][b2]; break;
	 case 6:
	    int_matrix[a1][b1] *= int_matrix[a2][b2]; break;
	 case 7:
	    if ((int_matrix[a2][b2] != 0) &&	   // Avoid divide by zero
		(int_matrix[a1][b1] < INT_MAX) && // For some reason in VC Debug mode I've
		// integer overflow error can occur when 
		// one value equals INT_MIN, and the other -1,
		// so disallow that computation
		(int_matrix[a1][b1] > INT_MIN) &&
		(int_matrix[a2][b2] < INT_MAX) &&
		(int_matrix[a2][b2] > INT_MIN)) 
	       int_matrix[a1][b1] /= int_matrix[a2][b2]; 
	    break;
	 } 
      }
   }
}
